# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

from metapack import Downloader
from metapack.cli.core import update_name, process_schemas, MetapackCliMemo, warn, prt, \
    update_schema_properties, write_doc

downloader = Downloader.get_instance()


def update(subparsers):
    parser = subparsers.add_parser(
        'update',
        help='Update a metatab file or metatpack package',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.set_defaults(run_command=run_update)

    parser.add_argument('metatabfile', nargs='?',
                        help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    parser.set_defaults(handler=None)

    ##
    ## Build Group

    group = parser.add_mutually_exclusive_group()

    group.add_argument('-s', '--schemas', default=False, action='store_true',
                       help='Rebuild the schemas for files referenced in the resource section')

    group.add_argument('-P', '--schema-properties', default=False, action='store_true',
                       help='Load schema properties from generators and upstream sources')

    group.add_argument('-A', '--alt-name', default=False, action='store_true',
                       help='Move AltNames to column name')

    group.add_argument('--clean-properties', default=False, action='store_true',
                       help='Remove unused columns in the schema, like AltName')

    group.add_argument('-c', '--categories', default=False, action='store_true',
                       help='Update categories, creating a new categories.csv metadata file')

    group.add_argument('-n', '--name', action='store_true', default=False,
                       help="Update the Name from the Datasetname, Origin and Version terms")

    group.add_argument('-D', '--descriptions', action='store_true', default=False,
                       help='Import descriptions for package references')

    parser.add_argument('-C', '--clean', default=False, action='store_true',
                        help='Clean schema before processing')

    parser.add_argument('-F', '--force', action='store_true', default=False,
                        help='Force the operation')


def run_update(args):
    m = MetapackCliMemo(args, downloader)

    if m.args.schemas:
        update_schemas(m)

    if m.args.schema_properties:
        update_schema_props(m)

    if m.args.clean_properties:
        clean_properties(m)

    if m.args.alt_name:
        move_alt_names(m)

    if m.args.categories:
        update_categories(m)

    elif m.mtfile_url.scheme == 'file' and m.args.name:
        update_name(m.mt_file, fail_on_missing=True, force=m.args.force)

    elif m.args.descriptions:
        update_descriptions(m)


def update_schemas(m):
    update_name(m.mt_file, fail_on_missing=False, report_unchanged=False)

    resource = m.get_resource()

    force = m.args.force

    if m.resource:
        if not resource:
            warn("Resource {} is not in metadata".format(m.resource))
        else:
            force = True

    process_schemas(m.mt_file, resource=m.resource, cache=m.cache, clean=m.args.clean, force=force)


def update_schema_props(m):
    doc = m.doc

    update_schema_properties(doc, force=m.args.force)

    write_doc(doc)


def clean_properties(m):
    doc = m.doc

    doc.clean_unused_schema_terms()

    write_doc(doc)


def update_descriptions(m):
    doc = m.doc

    for ref in doc.references():
        v = ref.find_first('Description')

        ref['Description'] = ref.resource.description

        print(ref.name, id(ref))
        print("Updated '{}' to '{}'".format(ref.name, ref.description))

    # print(m.doc.as_csv())

    for ref in doc.references():
        v = ref.find_first('Description')

        print(ref.name, id(ref), ref.description)

    write_doc(doc)


def update_categories(m):
    import metapack as mp

    doc = mp.MetapackDoc()
    doc['Root'].get_or_new_term('Root.Title', 'Schema and Value Categories')
    doc.new_section('Schema', ['Description', 'Ordered'])

    update_resource_categories(m, m.get_resource(), doc)

    doc.write_csv('categories.csv')


def update_resource_categories(m, resource, doc):
    columns = resource.row_generator.columns

    tab = doc['Schema'].new_term('Root.Table', m.get_resource().name)

    for col in columns:

        doc_col = tab.new_child('Column', col.get('name'), description=col.get('description'))

        if col.get('ordered'):
            doc_col.new_child('Column.Ordered', 'true')

        for k, v in col.get('values', {}).items():
            doc_col.new_child('Column.Value', k, description=v)

    doc.cleanse()  # Creates Modified and Identifier

    return doc


def move_alt_names(m):
    doc = m.doc

    for t in doc['Schema'].find('Root.Table'):
        moved = 0
        for c in t.find('Table.Column'):

            altname = c.get('AltName')
            if altname:
                c.name = altname.value
                c.remove_child(altname)

                moved += 1

        prt("Moved {} names in '{}'".format(moved, t.name))

    write_doc(doc)



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

import json
import yaml
from metapack import Downloader
from metapack.cli.core import find_csv_packages
from metapack.cli.core import prt, err, get_config, MetapackCliMemo
from metapack.jupyter.convert import convert_wordpress
from metapack.util import ensure_dir
from os.path import basename
from rowgenerators import parse_app_url
from textwrap import dedent
from uuid import uuid4

downloader = Downloader.get_instance()

def wp(subparsers):
    parser = subparsers.add_parser(
        'wp',
        help='Publish a Jupyter notebook to Wordpress ',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.set_defaults(run_command=run_wp)


    parser.add_argument('-g', '--group', help="Add an extra group")

    parser.add_argument('-H','--header', help='Dump YAML of notebook header', action='store_true')

    parser.add_argument('-G', '--get', help='Get and dump the post, rather that update it', action='store_true')

    parser.add_argument('-p', '--profile', help="Name of a BOTO or AWS credentails profile", required=False)

    parser.add_argument('site_name', help="Site name, in the .metapack.yaml configuration file")

    parser.add_argument('source', help="Path to a notebook file or a Metapack package")

def run_wp(args):

    u = parse_app_url(args.source)

    if u.target_format == 'ipynb':
        run_notebook(args)
    else:
        args.metatabfile = args.source
        m = MetapackCliMemo(args, downloader)

        run_package(m)

def run_notebook(args):

    p = '/tmp/metapack-wp-notebook/'
    ensure_dir(p)

    output_file, resources = convert_wordpress(args.source, p)

    r, post = publish_wp(args.site_name, output_file, resources, args)
    prt("Post url: ", post.link)

def get_site_config(site_name):

    config = get_config()

    if config is None:
        err("No metatab configuration found. Can't get Wordpress credentials. Maybe create '~/.metapack.yaml'")

    site_config = config.get('wordpress', {}).get(site_name, {})

    if not site_config:
        err("In config file '{}', expected 'wordpress.{}' section for site config"
            .format(config['_loaded_from'], site_name))

    if 'url' not in site_config or 'user' not in site_config or 'password' not in site_config:
        err(dedent(
            """
            Incomplete configuration. Expected:
                wordpress.{site_name}.url
                wordpress.{site_name}.user
                wordpress.{site_name}.password
            In configuration file '{cfg_file}'
            """.format(site_name=site_name, cfg_file=config['_loaded_from'])
        ))

    return site_config['url'], site_config['user'], site_config['password']

def prepare_image(slug, file_name, post_id):
    from xmlrpc import client as xmlrpc_client

    # prepare metadata
    data = {
        'name': 'picture.jpg',
        'type': 'image/jpeg',  # mimetype
    }

    # read the binary file and let the XMLRPC library encode it into base64
    with open(file_name, 'rb') as img:
        return {
            'name': '{}-{}'.format(slug,basename(file_name)),
            'type': 'image/png',  # mimetype
            'bits': xmlrpc_client.Binary(img.read()),
            'post_id': post_id
        }

def cust_field_dict(post):
    try:
        return dict( (e['key'],e['value']) for e in post.custom_fields)
    except (KeyError, AttributeError):
        return {}

def find_post(wp, identifier):

    from wordpress_xmlrpc.methods.posts import GetPosts

    for _post in wp.call(GetPosts()):
        if cust_field_dict(_post).get('identifier') == identifier:
            return  _post

    return None

def set_custom_field(post, key, value):

    if not hasattr(post, 'custom_fields') or not key in [ e['key'] for e in post.custom_fields]:

        post.custom_fields = [
            {'key': key, 'value': value},
        ]

def publish_wp(site_name, output_file, resources, args):

    """Publish a notebook to a wordpress post, using Gutenberg blocks.

    Here is what the metadata looks like, in a section of the notebook tagged 'frontmatter'

    show_input: hide
    github: https://github.com/sandiegodata/notebooks/blob/master/tutorial/American%20Community%20Survey.ipynb
    identifier: 5c987397-a954-46ca-8743-bdcd7a71579c
    featured_image: 171
    authors:
    - email: eric@civicknowledge.com
      name: Eric Busboom
      organization: Civic Knowledge
      type: wrangler
    tags:
    - Tag1
    - Tag2
    categories:
    - Demographics
    - Tutorial

    'Featured_image' is an attachment id

    """

    from wordpress_xmlrpc import Client, WordPressPost
    from wordpress_xmlrpc.methods.media import UploadFile, GetMediaLibrary
    from wordpress_xmlrpc.methods.posts import NewPost, EditPost, GetPost

    # http://busboom.org/wptest/wp-content/uploads/sites/7/2017/11/output_16_0-300x200.png

    url, user, password = get_site_config(site_name)

    meta = {}
    for r in resources:
        if r.endswith('.json'):
            with open(r) as f:
                meta = json.load(f)

    fm = meta.get('frontmatter',{})

    if not 'identifier' in fm or not fm['identifier']:
        err("Can't publish notebook without a unique identifier. Add this to the "
            "Metatab document or frontmatter metadata:\n   identifier: {}".format(str(uuid4())))

    wp = Client(url, user, password)

    post = find_post(fm['identifier'])

    if post:
        prt("Updating old post")
    else:
        post = WordPressPost()
        post.id = wp.call(NewPost(post))
        prt("Creating new post")

    post.title = fm.get('title','')
    post.slug = fm.get('slug')

    with open(output_file) as f:
        content = f.read()

    post.terms_names = {
        'post_tag': fm.get('tags',[]),
        'category':  fm.get('categories',[])
    }

    if args.header:
        print(yaml.dump(fm, default_flow_style=False))

    set_custom_field(post, 'identifier', fm['identifier'])

    post.excerpt = fm.get('excerpt', fm.get('brief', fm.get('description')))

    def strip_image_name(n):
        """Strip off the version number from the media file"""
        from os.path import splitext
        import re
        return re.sub(r'\-\d+$','',splitext(n)[0])

    extant_files = list(wp.call(GetMediaLibrary(dict(parent_id=post.id))))

    def find_extant_image(image_name):
        for img in extant_files:
            if strip_image_name(basename(img.metadata['file'])) == strip_image_name(image_name):
                return img

        return None

    for r in resources:

        image_data = prepare_image(fm['identifier'], r, post.id)
        img_from = "/{}/{}".format(fm['slug'], basename(r))

        extant_image = find_extant_image(image_data['name'])

        if  extant_image:
            prt("Post already has image:", extant_image.id, extant_image.link)
            img_to = extant_image.link

        elif r.endswith('.png'): # Foolishly assuming all images are PNGs
            response = wp.call(UploadFile(image_data, overwrite=True))

            prt("Uploaded image {} to id={}, {}".format(basename(r), response['id'], response['link']))

            img_to = response['link']

        content = content.replace(img_from, img_to)

    if fm['featured_image']:
        post.thumbnail = int(fm['featured_image'])
    elif isinstance(post.thumbnail, dict):
        # The thumbnail expects an attachment id on EditPost, but returns a dict on GetPost
        post.thumbnail = post.thumbnail['attachment_id']

    post.content = content

    r = wp.call(EditPost(post.id, post))

    return r,  wp.call(GetPost(post.id))

def html(doc):
    from markdown import markdown as convert_markdown
    from metapack.html import  markdown

    extensions = [
        'markdown.extensions.extra',
        'markdown.extensions.admonition'
    ]

    return convert_markdown(markdown(doc, template='wordpress.md'), extensions)


def run_package(m):
    from wordpress_xmlrpc import Client, WordPressPost
    from wordpress_xmlrpc.methods.posts import NewPost, EditPost

    print(m.doc.ref)

    url, user, password = get_site_config(m.args.site_name)

    doc = m.doc

    if not doc.find('Root.Distribution'):

        doc = find_csv_packages(m, downloader)
        prt("Package has no Root.Distribution, using CSV package: ", doc.ref)

    wp = Client(url, user, password)

    post = find_post(wp, doc.identifier)

    if m.args.get:
        if post:
            print(post.content)
            return

    if post:
        prt("Updating old post")
        action = lambda post: EditPost(post.id, post)
    else:
        prt("Creating new post")
        action = lambda post: NewPost(post)
        post = WordPressPost()


    set_custom_field(post, 'identifier', doc.identifier)

    post.excerpt = doc['Root'].get_value('Root.Description')

    post.terms_names = {
        'post_tag': [ t.value for t in doc['Root'].find('Root.Tag') ] + \
                    [t.value for t in doc['Root'].find('Root.Group')] + \
                    [doc['Root'].get_value('Root.Origin')],
        'category': ['Dataset']
    }

    post.title = doc.get_value('Root.Title')

    post.content = html(doc)

    r = wp.call(action(post))

    return r


# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# MIT License, included in this distribution as LICENSE

"""
CLI program for managing packages
"""

import sys

import requests
from metapack import Downloader, MetapackDoc
from metapack.cli.core import list_rr, warn, prt, err, MetapackCliMemo, write_doc

from metapack.jupyter.core import edit_notebook, set_cell_source
from metapack.util import ensure_dir
from os.path import dirname, basename, exists, splitext

downloader = Downloader.get_instance()


def notebook(subparsers):
    parser = subparsers.add_parser(
        'notebook',
        help='Create and convet Jupyter notebooks ',
        epilog='Cache dir: {}\n'.format(str(downloader.cache.getsyspath('/'))))

    parser.add_argument('notebook', nargs='?',
                        help="Path to a notebok file' ")

    parser.set_defaults(handler=None)

    ##
    ## New Notebooks

    cmdsp = parser.add_subparsers(help='sub-command help')

    cmdp = cmdsp.add_parser('new', help='Create new notebooks')
    cmdp.set_defaults(run_command=new_cmd)

    cmdp.add_argument('-m', '--metatab', action='store_true', default=False,
                      help='Create a metatab notebook from a metatab file')

    cmdp.add_argument('-E', '--eda', action='store_true', default=False,
                      help='Create an EDA notebook for a resource')

    cmdp.add_argument('-b', '--notebook', action='store_true', default=False,
                      help='Create a new, blank notebook')

    cmdp.add_argument('metatabfile', nargs='?',
                      help="Path or URL to a metatab file. If not provided, defaults to 'metadata.csv' ")

    ##
    ## Build Group

    cmdp = cmdsp.add_parser('convert', help='Convert to other formats')
    cmdp.set_defaults(run_command=convert_cmd)

    cmdp.add_argument('-H', '--hugo', default=False, nargs='?',
                      help='Write images and Markdown into a Hugo static site directory. or use '
                           'METAPACK_HUGO_DIR env var')

    cmdp.add_argument('-m', '--metatab', default=False,
                      help='Convert metatab data between .ipynb and .csv files, depending on argument')


def new_cmd(args):
    downloader = Downloader.get_instance()

    m = MetapackCliMemo(args, downloader)

    if m.args.eda:
        write_eda_notebook(m)

    elif m.args.notebook:
        write_notebook(m)

    elif m.args.metatab:
        write_metatab_notebook(m)


def convert_cmd(args):
    from metapack.jupyter.convert import convert_hugo

    if False:  # args.package:
        convert_notebook(args.notebook)

    elif False:  # args.documentation:
        convert_documentation(args.notebook)

    elif False:  # args.metatab:
        doc = extract_metatab(args.notebook)

        if args.lines:
            for line in doc.lines:
                print(": ".join(line))
        else:
            print(doc.as_csv())

    elif args.hugo:
        convert_hugo(args.notebook, args.hugo)

    elif args.metatab:
        convert_metatab_notebook(args.metatab)


def write_notebook(m):
    # Get the EDA notebook file from Github

    url = "https://raw.githubusercontent.com/Metatab/notebook-templates/master/package-notebook.ipynb"

    r = requests.get(url, allow_redirects=True)
    r.raise_for_status()

    nb_path = 'notebooks/new-notebook.ipynb'

    ensure_dir(dirname(nb_path))

    if exists(nb_path):
        err("Notebook {} already exists".format(nb_path))

    with open(nb_path, 'wb') as f:
        f.write(r.content)

    prt('Wrote {}'.format(nb_path))


def write_eda_notebook(m):
    # Get the EDA notebook file from Github

    url = "https://raw.githubusercontent.com/Metatab/exploratory-data-analysis/master/eda.ipynb"

    resource = m.get_resource()

    if not resource:
        warn('Must specify a resource. Select one of:')
        list_rr(m.doc)
        sys.exit(0)

    r = requests.get(url, allow_redirects=True)
    r.raise_for_status()

    nb_path = 'notebooks/{}-{}.ipynb'.format(splitext(basename(url))[0], resource.name)

    ensure_dir(dirname(nb_path))

    if exists(nb_path):
        err("Notebook {} already exists".format(nb_path))

    with open(nb_path, 'wb') as f:
        f.write(r.content)

    prt('Wrote {}'.format(nb_path))

    with edit_notebook(nb_path) as nb:
        set_cell_source(nb, 'resource_name', "resource_name='{}'".format(resource.name))


def write_metatab_notebook(m):
    from metapack.jupyter.convert import write_metatab_notebook as _write_metatab_notebook
    _write_metatab_notebook(m.doc)


def convert_metatab_notebook(source):
    from metapack.jupyter.convert import extract_notebook_metatab
    from metapack.jupyter.convert import write_metatab_notebook as _write_metatab_notebook
    from pathlib import Path

    source = Path(source)

    if source.suffix == '.csv':
        dest = source.with_suffix('.ipynb')
        doc = MetapackDoc(source)
        doc.ensure_identifier()
        doc.update_name(create_term=True)
        _write_metatab_notebook(doc, dest)

    elif source.suffix == '.ipynb':
        dest = source.with_suffix('.csv')

        doc = extract_notebook_metatab(source)
        doc.ensure_identifier()
        doc.update_name(create_term=True)
        write_doc(doc, dest)

    else:
        err("Source file must be either .ipynb or .csv")




from math import pi
from geoid.acs import Tract, AcsGeoid
from geoid.census import County, CensusGeoid
from geoid.tiger import Blockgroup, TigerGeoid
from random import randint, random
from datetime import datetime, date, time

def doubleit(v):

    return 2*v
    
    
def row_generator(resource, doc, env, *args, **kwargs):
    """ An example row generator function.

    Reference this function in a Metatab file as the value of a Datafile:

            Datafile: python:pylib#row_generator

    The function must yield rows, with the first being headers, and subsequenct rows being data.

    :param resource: The Datafile term being processed
    :param doc: The Metatab document that contains the term being processed
    :param args: Positional arguments passed to the generator
    :param kwargs: Keyword arguments passed to the generator
    :return:


    The env argument is a dict with these environmental keys:

    * CACHE_DIR
    * RESOURCE_NAME
    * RESOLVED_URL
    * WORKING_DIR
    * METATAB_DOC
    * METATAB_WORKING_DIR
    * METATAB_PACKAGE

    It also contains key/valu pairs for all of the properties of the resource.

    """

    yield 'int float prop ratio str acs_tract census_tract tiger_tract date time datetime'.split()

    for i in range(10):
        
        tract = Tract(6,72,randint(1,10))
        
        yield [
            i,
            float(i)/pi,
            random(),
            float(randint(1,1000))/float(randint(1,1000)),
            'string-'+str(i),
            tract,
            tract.convert(CensusGeoid),
            tract.convert(TigerGeoid),
            date(randint(2000,2020), randint(1,12), randint(1,27)),
            time(randint(0,23), randint(0,59), randint(0,59)),
            datetime(randint(2000,2020), randint(1,12), randint(1,27), randint(0,23), randint(0,59), randint(0,59))
            
            
        ]
        
        


# Copyright (c) 2016 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Create Markdown and HTML of datasets.
"""

import datetime
from rowgenerators import parse_app_url
from markdown import markdown as convert_markdown
from nameparser import HumanName
from pybtex import PybtexEngine
from pybtex.backends.html import Backend as HtmlBackend
from pybtex.style.formatting import toplevel
from pybtex.style.formatting.plain import Style
from pybtex.style.template import (
    words, together, field, sentence, optional_field, href, optional
)
from yaml import safe_dump

from metatab.doc import MetatabDoc

dl_templ = "{}\n:   {}\n\n"

def ns(v):
    """Return empty str if bool false"""
    return str(v) if bool(v) else ''


def linkify(v, description=None, cwd_url=None):


    if not v:
        return None

    u = parse_app_url(v)

    if u.scheme in ('http', 'https', 'mailto'):

        if description is None:
            description = v
        return '[{desc}]({url})'.format(url=v, desc=description)

    elif u.scheme == 'file':

        return '[{desc}]({url})'.format(url=u.path, desc=description)

    else:
        return v


def resource(r, fields=None):
    fields = fields or (
        ('Header', 'header'),
        ('Datatype', 'datatype'),
        ('Description', 'description')
    )

    headers = [f[0] for f in fields]
    keys = [f[1] for f in fields]

    rows = [''.join(["<td>{}</td>".format(e.replace("\n", "<br/>\n")) for e in [c.get(k, '') for k in keys]])
            for c in r.columns()]

    return ("### {name} \n [{url}]({url})\n\n".format(name=r.name, url=r.resolved_url)) + \
           "{}\n".format(ns(r.description)) + \
           "<table class=\"table table-striped\">\n" + \
           ("<tr>{}</tr>".format(''.join("<th>{}</th>".format(e) for e in headers))) + \
           "\n".join("<tr>{}</tr>".format(row) for row in rows) + \
           '</table>\n'


def ckan_resource_markdown(r, fields=None):
    fields = fields or (
        ('Header', 'header'),
        ('Datatype', 'datatype'),
        ('Description', 'description')
    )

    headers = [f[0] for f in fields]
    keys = [f[1] for f in fields]

    def col_line(c):
        return "* **{}** ({}): {}\n".format(
            c.get('header'),
            c.get('datatype'),
            "*{}*".format(c.get('description')) if c.get('description') else '')

    return "### {}. {} Columns. \n\n{}".format(
        ns(r.description),
        len(list(r.columns())),
        ''.join([col_line(c) for c in r.columns()])
    )


def resource_block(doc, fields=None):
    if fields is None:
        fields = [
            ('Header', 'header'),
            ('Datatype', 'datatype'),
            ('Description', 'description')
        ]

        for h in doc['Schema'].args:

            if h.lower() in ('note', 'notes', 'coding'):
                fields.append((h, h.lower()))

    return "".join(resource(r, fields) for r in doc['Resources'].find('Root.Resource'))


def resource_ref(r):
    return dl_templ.format(r.name, "_{}_<br/>{}".format(r.resolved_url, ns(r.description)))


def resource_ref_block(doc):
    return "".join(resource_ref(r) for r in doc['Resources'].find('Root.Resource'))


def contact(r):
    pass


class MetatabStyle(Style):
    # Minnesota Population Center. IPUMS Higher Ed: Version 1.0 [dataset]
    # Minneapolis, MN: University of Minnesota, 2016. http://doi.org/10.18128/D100.V1.0.

    def format_url(self, e):
        return words[
            href[
                field('url', raw=True),
                field('url', raw=True)
            ]
        ]

    def format_accessed(self, e):
        from dateutil.parser import parser

        return words[
            'Accessed',
            field('accessdate', raw=True, apply_func=lambda v: str(parser().parse(v).strftime("%d %b %Y")))
        ]

    def format_dataset(self, e):

        template = toplevel[
            optional[sentence[field('origin')]],
            self.format_btitle(e, 'title'),
            optional[sentence[together['Version', field('version')]]],
            optional[sentence[field('publisher') ]],
            optional[self.format_author_or_editor(e)],
            optional[words[optional_field('month'), field('year')]],
            self.format_web_refs(e),
            self.format_accessed(e)
        ]

        return template.format_data(e)


class MetatabHtmlBackend(HtmlBackend):
    def write_prologue(self):
        pass
        # super().write_prologue()

    def write_epilogue(self):
        pass
        # super().write_epilogue()

    def write_entry(self, key, label, text):
        self.output("<div class='citation'><a name=\"{key}\"><b>[{key}]</b></a> {text} </div>"
                    .format(key=key, text=text))


def make_citation_dict(td):
    """
    Update a citation dictionary by editing the Author field
    :param td: A BixTex format citation dict or a term
    :return:
    """

    from datetime import datetime

    if isinstance(td, dict):
        d = td
        name = d['name_link']
    else:

        d = td.as_dict()
        d['_term'] = td

        try:
            d['name_link'] = td.name
        except AttributeError:
            d['name_link'] = td['name_link'].value

    if 'author' in d and isinstance(d['author'], str):
        authors = []
        for e in d['author'].split(';'):
            author_d = HumanName(e).as_dict(include_empty=False)
            if 'suffix' in author_d:
                author_d['lineage'] = author_d['suffix']
                del author_d['suffix']
            authors.append(author_d)
        d['author'] = authors

    if not 'type' in d:

        if '_term' in d:
            t = d['_term']

            if t.term_is('Root.Reference') or t.term_is('Root.Resource'):
                d['type'] = 'dataset'

            elif t.term_is('Root.Citation'):
                d['type'] = 'article'
            else:
                d['type'] = 'article'

    if d['type'] == 'dataset':

        if not 'editor' in d:
            d['editor'] = [HumanName('Missing Editor').as_dict(include_empty=False)]

        if not 'accessdate' in d:
            d['accessdate'] = datetime.now().strftime('%Y-%m-%d')

    if not 'author' in d:
        d['author'] = [HumanName('Missing Author').as_dict(include_empty=False)]

    if not 'title' in d:
        d['title'] = d.get('description', '<Missing Title>')

    if not 'journal' in d:
        d['journal'] = '<Missing Journal>'

    if not 'year' in d:
        d['year'] = '<Missing Year>'

    if '_term' in d:
        del d['_term']

    return d


def make_metatab_citation_dict(t):
    """
    Return a dict with BibText key/values for metatab data
    :param t:
    :return:
    """

    try:

        if parse_app_url(t.url).proto == 'metatab':

            try:
                url = parse_app_url(str(t.resolved_url)).resource_url
                doc = t.row_generator.generator.package

            except AttributeError as e:
                return False  # It isn't a resource or reference

            creator = doc.find_first('Root.Creator')
            author_key='author'

            if not creator:
                creator = doc.find_first('Root.Wrangler')
                author_key = 'editor'

            if not creator:
                creator = doc.find_first('Root.Origin')
                author_key = 'editor'

            try:
                origin = doc['Contacts'].get('Root.Origin').get('organization').value
            except AttributeError:
                try:
                    origin = doc.get_value('Root.Origin', doc.get_value('Name.Origin')).title()
                except:
                    origin = None

            d = {
                'type': 'dataset',
                'name_link': t.name,
                author_key: [HumanName(creator.value).as_dict(include_empty=False)],
                'publisher': creator.properties.get('organization'),
                'origin': origin,
                'journal': '2010 - 2015 American Community Survey',
                'title': doc['Root'].find_first_value('Root.Title') + '; ' + t.name,
                'year': doc.get_value('Root.Year'),
                'accessDate': '{}'.format(datetime.datetime.now().strftime('%Y-%m-%d')),
                'url': url,
                'version': doc.get_value('Root.Version', doc.get_value('Name.Version') ),
            }

            d =  { k:v for k, v in d.items() if v is not None}


            return d


        else:
            return False

    except (AttributeError, KeyError) as e:
        raise
        return False


def _bibliography(doc, terms, converters=[], format='html'):
    """
    Render citations, from a document or a doct of dicts

    If the input is a dict, each key is the name of the citation, and the value is a BibTex
    formatted dict

    :param doc: A MetatabDoc, or a dict of BibTex dicts
    :return:
    """

    output_backend = 'latex' if format == 'latex' else MetatabHtmlBackend

    def mk_cite(v):
        for c in converters:
            r = c(v)

            if r is not False:
                return r

        return make_citation_dict(v)

    if isinstance(doc, MetatabDoc):
        # This doesn't work for LaTex, b/c the formatter adds the prologue and epilogue to eery entry

        d = [mk_cite(t) for t in terms]

        cd = {e['name_link']: e for e in d}

    else:

        cd = {k: mk_cite(v, i) for i, (k, v) in enumerate(doc.items())}

    # for k, v in cd.items():
    #    print (k, v)

    return PybtexEngine().format_from_string(safe_dump({'entries': cd}),
                                             style=MetatabStyle,
                                             output_backend=output_backend,
                                             bib_format='yaml')


def bibliography(doc, converters=[], format='html'):
    terms = list(doc.get_section('Bibliography', []))

    return _bibliography(doc, terms, converters, format)


def data_sources(doc, converters=[], format='html'):
    terms = list(doc.references()) + list(doc.resources())

    return _bibliography(doc, terms, converters, format)


def identity_block(doc):
    from collections import OrderedDict

    name = doc.find_first('Root.Name')

    d = OrderedDict(
        name=name.value,
        identifier=doc.find_first_value('Root.Identifier'),
    )

    for k, v in name.props.items():
        d[k] = v

    try:
        del d['@value']
    except KeyError:
        pass

    return "".join(dl_templ.format(k, v) for k, v in d.items())


def modtime_str(doc):
    from dateutil.parser import parse

    modtime = doc.find_first_value('Root.Modified') or doc.find_first_value('Root.Issued')

    if modtime:
        try:
            modtime_str = "Last Modified: " + parse(modtime).strftime("%Y-%m-%d %H:%M")
        except ValueError:
            modtime_str = ''

    else:

        modtime_str = ''

    return modtime_str



def process_contact(d):

    email = d.get('email',None)
    org = d.get('organization', None)
    url = d.get('url',None)
    name = d.get('name', None)

    if name and email and org and url:
        return {
            'parts':["[{}]({})".format(name,'mailto:'+email),"[{}]({})".format(org,url)]
        }

    elif name and email and org:
        return {
            'parts':["[{}]({})".format(name, 'mailto:' + email),"{}".format(org)]
        }
    elif name and email and url:
        return {
            'parts': ["[{}]({})".format(name, 'mailto:' + email), "{}".format(url)]
        }

    elif name and org and url:
        return {
            'parts':["{}".format(name),"[{}]({})".format(org, url)]
        }

    elif name and org:
        return {
            'parts': ["{}".format(name),"{}".format(org)]
        }
    elif name and url:
        return {
            'parts': ["{}".format(name),"[{}]({})".format(url, url)]
        }
    elif name:
        return {
            'parts': ["{}".format(name)]
        }
    elif org and email and url:
        return {
            'parts': ["[{}]({})".format(org,url),"[{}]({})".format(email,'mailto:'+email)]
        }
    elif org and email:
        return {
            'parts':["{}".format(org),"[{}]({})".format(email, 'mailto:' + email)]
        }
    elif org and url:
        return {
            'parts':[ "[{}]({})".format(org, url)]
        }
    elif url and email:
        return {
            'parts':["[{}]({})".format(url, url), "[{}]({})".format(email, 'mailto:' + email)]
        }

    elif org:
        return {
            'parts': ["{}".format(org)]
        }
    elif url:
        return {
            'parts': ["[{}]({})".format(url, url)]
        }
    elif email:
        return {
            'parts': ["[{}]({})".format(email, 'mailto:' + email)]
        }
    else:

        return {
            'parts': []
        }

def display_context(doc):
    """Create a Jinja context for display"""
    from rowgenerators.exceptions import DownloadError

    context = {s.name.lower(): s.as_dict() for s in doc if s.name.lower() != 'schema'}

    #import json
    #print(json.dumps(context, indent=4))

    mandatory_sections = ['documentation', 'contacts']

    # Remove section names
    deletes = []
    for k,v in context.items():
        try:
            del v['@value']
        except KeyError:
            pass # Doesn't have the value
        except TypeError:
            # Is actually completely empty, and has a scalar value. Delete and re-create
            deletes.append(k)

        if isinstance(v, str): # Shouldn't ever happen, but who knows ?
            deletes.append(k)

    for d in deletes:
        try:
            del context[d]
        except KeyError:
            # Fails in TravisCI, no idea why.
            pass

    for ms in mandatory_sections:
        if not ms in context:
            context[ms] = {}

    # Load inline documentation
    inline = ''

    for d in context.get('documentation',{}).get('documentation',[]):

        u = parse_app_url(d['url'])

        if u.target_format == 'md':  # The README.md file
            inline = ''
            if u.proto == 'file':
                # File really ought to be relative
                t = doc.package_url.join_target(u).get_resource().get_target()
            else:
                try:
                    t = u.get_resource().get_target()
                except DownloadError as e:
                    raise e


            try:
                with open(t.fspath) as f:
                    inline += f.read()

            except FileNotFoundError:
                pass

            del d['title'] # Will cause it to be ignored in next section

    # Strip off the leading title, if it exists, because it will be re-applied
    # by the templates
    import re

    lines = inline.strip().splitlines()
    if lines and lines[0].startswith('# '):
        lines = lines[1:]

    context['inline_doc'] = '\n'.join(lines)

    # Convert doc section
    doc_links = {}
    images = {}
    for term_name, terms in context['documentation'].items():
        if term_name == 'note':
            context['notes'] = terms
        else:
            for i, term in enumerate(terms):
                try:
                    if term_name == 'image':
                        images[term['title']] = term
                    else:
                        doc_links[term['title']] = term
                except AttributeError: # A scalar
                    pass # There should not be any scalars in the documentation section
                except KeyError:
                    pass # ignore entries without titles
                except TypeError:
                    pass # Also probably a ascalar

    context['doc_links'] = doc_links
    context['images'] = images

    del context['documentation']

    #
    # Update contacts

    origin = None
    for term_name, terms in context['contacts'].items():
        if isinstance(terms, dict):
            origin = terms # Origin is a scalar in roort, must be converted to sequence here
        else:
            for t in terms:
                try:
                    t.update(process_contact(t))
                except AttributeError:
                    pass # Probably got a scalar
    if origin:
        origin.update(process_contact(origin))
        context['contacts']['origin'] = [origin]


    # For resources and references, convert scalars into lists of dicts, which are the
    # default for Datafiles and References.

    for section in ('references', 'resources'):
        for term_key, term_vals in context.get(section,{}).items():
            if isinstance(term_vals, dict):
                if '@value' in term_vals:
                    term_vals['url'] = term_vals['@value']
                    del term_vals['@value']
                new_term_vals = [term_vals]
            elif isinstance(term_vals, list):
                new_term_vals = None
            else:
                new_term_vals = [ {'url': term_vals, 'name': term_vals}]

            if new_term_vals:
                context[section][term_key] = new_term_vals

    context['distributions'] = {}
    for dist in doc.find('Root.Distribution'):
        context['distributions'][dist.type] = dist.value

    if doc.find('Root.Giturl'):
        context['distributions']['source'] = doc.get_value('Root.Giturl')

    return context

def markdown(doc, title=True, template='short_documentation.md'):
    """Markdown, specifically for the Notes field in a CKAN dataset"""

    from jinja2 import Environment, PackageLoader, select_autoescape
    env = Environment(
        loader=PackageLoader('metapack', 'support/templates')
        #autoescape=select_autoescape(['html', 'xml'])
    )

    context = display_context(doc)

    return env.get_template(template).render(**context)


def html(doc, template='short_documentation.md'):
    extensions = [
        'markdown.extensions.extra',
        'markdown.extensions.admonition'
    ]

    return """
<html>
<head>
    <meta charset="utf-8">
    <meta http-equiv="X-UA-Compatible" content="IE=edge">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <!-- Latest compiled and minified CSS -->
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">

    <!-- Optional theme -->
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap-theme.min.css" integrity="sha384-rHyoN1iRsVXV4nD0JutlnGaslCJuC7uwjduW9SVrLvRYooPp2bWYgmgJQIXwl/Sp" crossorigin="anonymous">

    <style>
    .narrow-dl .dl-horizontal dd {{
        margin-left: 100px;
    }}
    
    img[alt=doc_img] {{ width: 300px; }}

    .narrow-dl .dl-horizontal dt {{
        width: 80px;
    }}

    </style>
    <title>{title}</title>
</head>
<body>
    <div class="container">
        <div class="row">
            {body}
        </div>
    </div>
</body>
</html>
    """.format(
        title=doc.find_first_value('Root.Title'),
        body=convert_markdown(markdown(doc, template=template), extensions)
    )



import unittest

def test_data(*paths):
    from os.path import dirname, join, abspath
    import metapack.test.test_data

    return join(dirname(abspath(metapack.test.test_data.__file__)), *paths)


def cache_fs():

    from fs.tempfs import TempFS

    return TempFS('rowgenerator')

def get_cache():
    return cache_fs()

class MetapackTest(unittest.TestCase):
    """Test Metapack AppUrls and Row Generators"""

    def setUp(self):
        import warnings
        warnings.simplefilter('ignore')



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
CLI program for storing pacakges in CKAN
"""

import os
from botocore.exceptions import NoCredentialsError
from metapack import MetapackDoc, MetapackUrl
from metapack import MetapackPackageUrl
from metapack.cli.core import prt, err, generate_packages, write_doc
from metapack.constants import PACKAGE_PREFIX
from metapack.index import SearchIndex, search_index_file
from metapack.package import *
from metapack.package.s3 import S3Bucket
from metapack.util import datetime_now
from metatab import DEFAULT_METATAB_FILE
from os import getcwd, getenv
from os.path import basename
from rowgenerators import parse_app_url
from rowgenerators.util import fs_join as join
from tabulate import tabulate


class MetapackCliMemo(object):

    def __init__(self, args):

        self.cwd = getcwd()

        self.args = args

        self.downloader = Downloader.get_instance()

        self.cache = self.downloader.cache

        self.mtfile_arg = self.args.metatabfile if self.args.metatabfile else join(self.cwd, DEFAULT_METATAB_FILE)

        self.mtfile_url = MetapackUrl(self.mtfile_arg, downloader=self.downloader)

        self.resource = self.mtfile_url.fragment

        self.package_url = self.mtfile_url.package_url
        self.mt_file = self.mtfile_url.metadata_url

        self.package_root = self.package_url.join(PACKAGE_PREFIX)

        if not self.args.s3:
            doc = MetapackDoc(self.mt_file)
            self.args.s3 = doc['Root'].find_first_value('Root.S3')

        self.s3_url = parse_app_url(self.args.s3)

        if self.s3_url and not self.s3_url.scheme == 's3':
            self.s3_url = parse_app_url("s3://{}".format(self.args.s3))

        self.doc = MetapackDoc(self.mt_file)

        access_value = self.doc.find_first_value('Root.Access')

        self.acl = 'private' if access_value == 'private' else 'public-read'

        self.bucket = S3Bucket(self.s3_url, acl=self.acl , profile=self.args.profile) if self.s3_url else None


def s3(subparsers):

    parser = subparsers.add_parser(
        's3',
        help='Create packages and store them in s3 buckets',
    )

    parser.set_defaults(run_command=run_s3)

    parser.add_argument('-p', '--profile', help="Name of a BOTO or AWS credentials profile", required=False)

    parser.add_argument('-s', '--s3', help="URL to S3 where packages will be stored", required=False)

    parser.add_argument('-F', '--force', default=False, action='store_true',
                       help='Force write for all files')

    parser.add_argument('-C', '--credentials', help="Show S3 Credentials and exit. "
                                                    "Eval this string to setup credentials in other shells.",
                        action='store_true', default=False)

    parser.add_argument('metatabfile', nargs='?', help='Path to a Metatab file')


def run_s3(args):

    m = MetapackCliMemo(args)

    if m.args.credentials:
        show_credentials(m.args.profile)
        exit(0)

    dist_urls, fs_p = upload_packages(m)

    if dist_urls:

        # Create the CSV package, with links into the filesystem package
        if fs_p:
            create_s3_csv_package(m, dist_urls, fs_p)
        else:
            # If this happens, then no packages were created, because an FS package
            # is always built first
            prt("Not creating CSV package; no FS package was uploaded")

    if dist_urls:

        rows = [ [path,url, reason] for  what, reason, url, path in fs_p.files_processed if what != 'skip']
        if rows:
            prt("\nWrote these files:")
            prt(tabulate(rows, headers='path url reason'.split()))

        rows = [[path,url, reason] for  what, reason, url, path in fs_p.files_processed if what == 'skip']
        if rows:
            prt("\nSkipped these files:")
            prt(tabulate(rows, headers='path url reason'.split()))

        prt("\nSynchronized these Package Urls")
        prt("-------------------------------")
        for au in dist_urls:
            prt(au)
        prt("-------------------------------")

    else:
        prt("Did not find any packages to upload")

    m.doc['Root'].get_or_new_term('Root.Issued').value = datetime_now()

    clear_cache(m, fs_p.files_processed)

def clear_cache(m, files_processed):
    """Remove any files we may have uploaded from the cache. """

    for what, reason, url, path in files_processed:
        cp = m.doc.downloader.cache_path(url)

        if m.cache.exists(cp):
            m.cache.remove(cp)

def add_to_index(p):

    idx = SearchIndex(search_index_file())

    idx.add_package(p, format='web')

    idx.write()

def set_distributions(doc, dist_urls):

    for t in doc.find('Root.Distribution'):
        doc.remove_term(t)

    for au in dist_urls:
        doc['Root'].new_term('Root.Distribution', au)

    path = write_doc(doc)()


def make_s3_package(file, package_root, cache, env, skip_if_exists, acl='public-read'):
    from metapack import MetapackUrl
    from metapack.package import S3PackageBuilder

    assert package_root

    p = S3PackageBuilder(file, package_root, callback=prt, env=env, acl=acl)

    if not p.exists() or not skip_if_exists:
        url = p.save()
        prt("Packaged saved to: {}".format(url))
        created = True
    elif p.exists():
        prt("S3 Filesystem Package already exists")
        created = False
        url = p.access_url

    return p, MetapackUrl(url, downloader=file.downloader), created


def upload_packages(m):
    """"""
    dist_urls = []
    fs_p = None

    files_processed = []

    # For each package in _packages with the same name as this document...
    for ptype, purl, cache_path in generate_packages(m):

        au = m.bucket.access_url(cache_path)

        # Just copy the Excel and Zip files directly to S3
        if ptype in ('xlsx', 'zip'):
            with open(purl.path, mode='rb') as f:
                access_url = m.bucket.write(f.read(), basename(purl.path), m.acl)

                if m.bucket.last_reason:
                    files_processed.append([*m.bucket.last_reason, access_url, '/'.join(purl.path.split(os.sep)[-2:])])

                prt("Added {} distribution: {} ".format(ptype, au))
                dist_urls.append(au)

        elif ptype == 'fs':  # Make the S3 package from the filesystem package

            env = {}
            skip_if_exist = False

            try:
                s3_package_root = MetapackPackageUrl(str(m.s3_url), downloader=m.downloader)

                fs_p, fs_url, created = make_s3_package(purl.metadata_url, s3_package_root, m.cache, env, skip_if_exist,
                                                        m.acl)
            except NoCredentialsError:
                print(getenv('AWS_SECRET_ACCESS_KEY'))
                err("Failed to find boto credentials for S3. "
                    "See http://boto3.readthedocs.io/en/latest/guide/configuration.html ")

            # A crappy hack. make_s3_package should return the correct url
            if fs_p:
                if m.acl == 'private':
                    au = fs_p.private_access_url.inner
                else:
                    au = fs_p.public_access_url.inner

                dist_urls.append(au)

    fs_p.files_processed += files_processed # Ugly encapsulating-breaking hack.

    return dist_urls, fs_p

def create_s3_csv_package(m, dist_urls, fs_p):

    u = MetapackUrl(fs_p.access_url, downloader=m.downloader)

    resource_root = u.dirname().as_type(MetapackPackageUrl)

    p = CsvPackageBuilder(u, m.package_root, resource_root)

    access_url = m.bucket.access_url(p.cache_path)
    dist_urls.append(access_url)

    for au in dist_urls:
        if not p.doc.find_first('Root.Distribution', str(au)):
            p.doc['Root'].new_term('Root.Distribution', au)

    # Re-write the URLS for the datafiles
    for r in p.datafiles:
        r.url = fs_p.bucket.access_url(r.url)

    # Rewrite Documentation urls:
    for r in p.doc.find(['Root.Documentation', 'Root.Image']):

        url = parse_app_url(r.url)
        if url.proto == 'file':
            r.url = fs_p.bucket.access_url(url.path)

    csv_url = p.save()

    with open(csv_url.path, mode='rb') as f:
        m.bucket.write(f.read(), csv_url.target_file, m.acl)


    add_to_index(open_package(access_url))

    if m.bucket.last_reason:
        # Ugly encapsulation-breaking hack.
        fs_p.files_processed += [ [*m.bucket.last_reason, access_url, '/'.join(csv_url.path.split(os.sep)[-2:])] ]

    # Create an alternative url with no version number, so users can get the
    # most recent version
    csv_non_ver_url = csv_url.join_dir("{}.{}".format(m.doc.nonver_name, csv_url.target_format))

    with open(csv_url.path, mode='rb') as f:
        m.bucket.write(f.read(), csv_non_ver_url.target_file, m.acl)

    s3_path = csv_non_ver_url.path.split(os.sep)[-1]

    access_url = m.bucket.access_url(s3_path)

    dist_urls.append(access_url)

    if m.bucket.last_reason:
        # Ugly encapsulation-breaking hack.
        fs_p.files_processed += [ [*m.bucket.last_reason, access_url, s3_path] ]



def show_credentials(profile):
    import boto3

    session = boto3.Session(profile_name=profile)

    if profile:
        cred_line = " 'eval $(metasync -C -p {} )'".format(profile)
    else:
        cred_line = " 'eval $(metasync -C)'"

    prt("export AWS_ACCESS_KEY_ID={} ".format(session.get_credentials().access_key))
    prt("export AWS_SECRET_ACCESS_KEY={}".format(session.get_credentials().secret_key))
    prt("# Run {} to configure credentials in a shell".format(cred_line))





# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""
Metapack CLI program for creating new metapack package directories
"""

import argparse
import requests
from metapack.jupyter.convert import write_metatab_notebook
from metapack.package import *
from pathlib import Path
from shutil import copyfile
from .core import MetapackCliMemo as _MetapackCliMemo
from metapack.jupyter.core import edit_notebook, set_cell_source, get_cell_source


downloader = Downloader.get_instance()

TEMPLATE_NOTEBOOK = "https://raw.githubusercontent.com/Metatab/exploratory-data-analysis/master/metadata.ipynb"


class MetapackCliMemo(_MetapackCliMemo):

    def __init__(self, args, downloader):
        super().__init__(args, downloader)


def new_args(subparsers):
    """
    The `mp new` command creates source package directories
    with a proper name, a `.gitignore` file, and optionally, example data,
    entries and code. Typical usage, for creating a new package with most
    of the example options, is ::
    
        mp new -o metatab.org -d tutorial -L -E -T "Quickstart Example Package" 
    
    The :option:`-C` option will set a configuration file, which is a
    Metatab file that with terms that are copied into the `metadata.csv` file
    of the new package. Currently, it copies a limited number of terms,
     including:
    
    - Terms in the Contacts section
    - Root.Space
    - Root.Time
    - Root.Grain
    - Root.Variant
    - Root.Version

    """
    parser = subparsers.add_parser(
        'new',
        help='Create new Metatab packages',
        description=new_args.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.set_defaults(run_command=new_cmd)

    parser.add_argument('-o', '--origin', help="Dataset origin, usually a second-level domain name. Required")
    parser.add_argument('-d', '--dataset', help="Main dataset name. Required", required=True)
    parser.add_argument('-t', '--time', help="Temporal extents, usually a year, ISO8601 time, or interval. ")
    parser.add_argument('-s', '--space', help="Space, geographic extent, such as a name of a state or a Census geoid")
    parser.add_argument('-g', '--grain', help="Grain, the type of entity a row represents")
    parser.add_argument('-v', '--variant', help="Variant, any distinguishing string")
    parser.add_argument('-r', '--revision', help="Version, defaults to 1", default=1)

    parser.add_argument('-T', '--title', help="Set the title")

    parser.add_argument('-L', '--pylib', help="Configure a pylib directory for Python code extensions",
                        action='store_true')

    parser.add_argument('-E', '--example', help="Add examples of resources",
                        action='store_true')

    parser.add_argument('-J', '--jupyter', help="Create a Jupyter notebook source package",
                        action='store_true')

    parser.add_argument('--template', help="Metatab file template, defaults to 'metatab' ", default='metatab')
    parser.add_argument('-C', '--config', help="Path to config file. "
                                               "Defaults to ~/.metapack-defaults.csv or value of METAPACK_DEFAULTS env var."
                                               "Sets defaults for specia root terms and the Contacts section.")
    return parser


def doc_parser():
    from .mp import base_parser

    return new_args(base_parser())

class EmptyTerm(object):

    value = None

et = EmptyTerm()

def new_cmd(args):
    from metapack import MetapackDoc
    from metapack.util import make_metatab_file, datetime_now, ensure_dir
    from metapack.cli.core import write_doc, prt, err
    from os.path import exists, join, expanduser, dirname
    from metatab import DEFAULT_METATAB_FILE
    from os import getenv
    import metapack.support as support_dir

    if args.config:
        config_file = args.config
    elif getenv("METAPACK_CONFIG"):
        config_file = getenv("METAPACK_DEFAULTS")
    elif expanduser('~/.metapack-default.csv'):
        config_file = expanduser('~/.metapack-defaults.csv')
    else:
        config_file = None

    if config_file and exists(config_file):
        prt(f"Using defaults file '{config_file}'")
        config = MetapackDoc(config_file)
    else:
        config = MetapackDoc()

    if args.jupyter:
        import tempfile

        with tempfile.NamedTemporaryFile(suffix='.ipynb', delete=False) as fp:

            r = requests.get(TEMPLATE_NOTEBOOK, allow_redirects=True)
            r.raise_for_status()
            nb_contents = r.content

            fp.write(r.content)
            nb_path = Path(fp.name)

        doc = MetapackDoc(str(nb_path))


    else:

        doc = make_metatab_file(args.template)

    doc['Root']['Created'] = datetime_now()

    origin = args.origin or config.get_value('Root.Origin')

    if not origin:
        err("Must specify a value for origin, either on command line or in defaults file")

    (doc['Root'].find_first('Root.Origin') or et).value = origin
    (doc['Root'].find_first('Root.Dataset') or et).value = args.dataset
    (doc['Root'].find_first('Root.Space') or et).value = args.space or config.get_value('Root.Space')
    (doc['Root'].find_first('Root.Time') or et).value = args.time or config.get_value('Root.Time')
    (doc['Root'].find_first('Root.Grain') or et).value = args.grain or config.get_value('Root.Grain')
    (doc['Root'].find_first('Root.Variant') or et).value = args.variant or config.get_value('Root.Variant')
    (doc['Root'].find_first('Root.Version') or et).value = args.revision or config.get_value('Root.Version')

    # Copy contacts in
    if 'Contacts' in config:
        for c in config['Contacts']:
            doc['Contacts'].add_term(c)

    if args.title:
        doc['Root'].find_first('Root.Title').value = args.title.strip()

    nv_name = doc.as_version(None)


    if args.example:
        doc['Resources'].new_term('Root.Datafile',
                                  'http://public.source.civicknowledge.com/example.com/sources/random-names.csv',
                                  name='random_names')

        doc['Documentation'].new_term('Root.Homepage', 'http://metatab.org', title='Metatab Home Page')

    doc.ensure_identifier()
    doc.update_name(create_term=True)

    if args.jupyter:

        new_nb_path = Path(f'{nv_name}.ipynb')

        doc['Resources'].new_term('Root.Datafile','./'+str(new_nb_path)+"#df" ,
                                  name='local_dataframe', description='Example of using a local Dataframe')

        if new_nb_path.exists():
            err(f"Directory {nb_path} already exists")

        copyfile(nb_path, new_nb_path)

        write_metatab_notebook(doc, new_nb_path)

        with edit_notebook(new_nb_path) as nb:
            init = get_cell_source(nb, 'init')
            init += f"\nthis_package_name = '{str(new_nb_path.name)}'"
            set_cell_source(nb,'init',init)
         
        nb_path.unlink()
    else:

        doc['Documentation'].new_term('Root.Documentation', 'file:README.md', title='README')

        if True:  # args.pylib:
            from metapack.support import pylib
            pylib_dir = join(nv_name, 'pylib')
            ensure_dir(pylib_dir)
            with open(join(pylib_dir, '__init__.py'), 'w') as f_out, open(pylib.__file__) as f_in:
                f_out.write(f_in.read())

            if args.example:
                doc['Resources'].new_term('Root.Datafile', 'python:pylib#row_generator', name='row_generator')

        if exists(nv_name):
            err(f"Directory {nv_name} already exists")

        ensure_dir(nv_name)

        prt(f"Writing to '{nv_name}'")

        write_doc(doc, join(nv_name, DEFAULT_METATAB_FILE))

        with open(join(dirname(support_dir.__file__), 'gitignore')) as f:
            gitignore = f.read()

        with open(join(nv_name, '.gitignore'), 'w') as f:
            f.write(gitignore)

        if args.title:
            readme = '# {}\n'.format(args.title)
        else:
            readme = '# {}\n'.format(doc.get_value('Root.Name'))

        with open(join(nv_name, 'README.md'), 'w') as f:
            f.write(readme)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from setuptools import find_packages, setup

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    sys.exit()

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme = f.read()

classifiers = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.4',
    'Topic :: Software Development :: Libraries :: Python Modules',
]


# noinspection PyTypeChecker
setup_d = dict(
    name='metapack',
    version='0.8.35',
    description='Data packaging system using Metatab',
    long_description=readme,
    packages=find_packages(),
    # package_data= {}, package_data is only used for binary distributions! ARRRG!

    zip_safe=False,
    install_requires=[
        'urllib3<1.24,>=1.20', # To keep botocore happy.
        'click<7.0,>=3.3', # For jsontableschema
        'python-dateutil<2.7.0', # Stupid botocode
        'boto3',
        'openpyxl<2.5', # Required by tabulator
        'unicodecsv',
        'pyyaml',
        'datapackage<1.0',
        'bs4',
        'markdown==2.6.11',
        'nbconvert',
        'IPython',
        'ipykernel',
        'jupyter',
        'nameparser',
        'pybtex',
        'rowgenerators>=0.7.16',
        'metatabdecl>=1.0.0',
        'metatab>=0.6.6',
        'tableintuit>=0.0.6',
        'geoid>=1.0.4',
        'terminaltables',
        'docopt',
        'jinja2'


        # 'wordpress_xmlrpc'# For `mp notebook -w`, sending notebooks to wordpress
    ],

    entry_points={
        'console_scripts': [
            'mp=metapack.cli.mp:mp',
        ],
        'nbconvert.exporters': [
            #'metapack = metapack.jupyter:MetapackExporter',
            #'hugo = metapack.jupyter:HugoExporter',
        ],

        'appurl.urls': [
            "metapack+ = metapack.appurl:MetapackUrl",
            ".ipynb = metapack.appurl:JupyterNotebookUrl",
            "index: = metapack.appurl:SearchUrl"

        ],
        'rowgenerators': [
            "<JupyterNotebookUrl> = metapack.rowgenerator:JupyterNotebookSource",
            "metapack+.txt =  metatab.rowgenerators:TextRowGenerator",
            "metatab+.ipynb =  metapack.rowgenerator:IpynbRowGenerator",
            "metapack+.ipynb =  metapack.rowgenerator:IpynbRowGenerator",

        ],
        'mt.subcommands': [
            'url=metapack.cli.url:url',
            'update=metapack.cli.update:update',
            'build=metapack.cli.build:build',
            's3=metapack.cli.s3:s3',
            'index=metapack.cli.index:index_args',
            'new=metapack.cli.new:new_args',
            'ckan=metapack.cli.metakan:metakan',
            'notebook=metapack.cli.notebook:notebook',
            'run=metapack.cli.run:run',
            'search=metapack.cli.search:search',
            'info=metapack.cli.info:info_args',
            'doc=metapack.cli.doc:doc_args',
            'stats=metapack.cli.stats:stats_args',
            'edit=metapack.cli.edit:edit_args',
            'open=metapack.cli.open:open_args',
            'wp=metapack.cli.wp:wp',
        ]
    },

    include_package_data=True,
    author='Eric Busboom',
    author_email='eric@civicknowledge.com',
    url='https://github.com/Metatab/metapack.git',
    license='BSD',
    classifiers=classifiers,
    extras_require={
        'test': ['datapackage'],
        'geo': ['fiona', 'shapely', 'pyproj', 'geopandas'],
        'jupyter': ['jupyter', 'pandas', 'geopandas' ],

    },

    test_suite='metapack.test.test_suite.suite',
    tests_require=['nose','publicdata', 'geopandas', 'fiona', 'shapely', 'pyproj', 'jupyter'],

)

setup(
    **setup_d
)



# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""

from csv import reader
from metapack.jupyter.core import get_cell_source
from metapack.jupyter.exec import execute_notebook
from metapack.util import get_materialized_data_cache
from metatab.rowgenerators import TextRowGenerator
from os.path import join, exists
from rowgenerators import Source, SourceError
from rowgenerators.exceptions import AppUrlError


# The class used to be here, and still gets referenced here

class JupyterNotebookSource(Source):
    """Generate rows from an IPython Notebook.

     This generator will execute a Jupyter notebook. Before execution, it adds an "%mt_materalize"
     magic to the notebook, which will cause the target dataframe to be written to a temporary file, and
     the temporary file is yielded to the caller. Not most efficient, but it fits the model.
     """

    def __init__(self, ref, cache=None, working_dir=None, env=None, doc=None, **kwargs):
        super().__init__(ref, cache, working_dir, **kwargs)

        self.env = env
        self.doc = doc

    def start(self):
        pass

    def finish(self):
        pass

    def __iter__(self):

        self.start()

        dr_name = get_materialized_data_cache(self.doc)

        path = join(dr_name, self.ref.target_file + ".csv")

        # The first python notebook in the resources will get executed, cache the datafiles,
        # then the remaining ones will just use the caches data.
        if not exists(path):

            # The execute_notebook() function will add a cell with the '%mt_materialize' magic,
            # with a path that will case the file to be written to the same location as
            # path, below.

            if not self.ref.target_dataframe():
                raise AppUrlError('Url did not specify a dataframe; use the "#" fragment ')

            nb = execute_notebook(self.ref.path, dr_name, [self.ref.target_dataframe()], True, self.env)

        with open(path) as f:
            yield from reader(f)

        self.finish()


def copy_reference(resource, doc, env, *args, **kwargs):
    """A row-generating function that yields from a reference. This permits an upstream package to be
    copied and modified by this package, while being formally referenced as a dependency

    The function will generate rows from a reference that has the same name as the resource term
    """

    yield from doc.reference(resource.name)


class IpynbRowGenerator(TextRowGenerator):
    """Generate metatab rows from the text lines in a Jupyter notebook"""

    def __init__(self, ref, cache=None, working_dir=None, path=None, **kwargs):


        super().__init__(ref, cache, working_dir, path, **kwargs)

        self._ref = ref

        self._path = path or '<none>'

    def _open(self):
        import nbformat

        ref = self._ref

        while True:

            try:
                # Pathlib Path
                with ref.open() as f:
                    nb = nbformat.read(f, as_version=4)

                break
            except:
                pass

            try:
                # Filehandle
                nb = nbformat.read(ref, as_version=4)
                break
            except:
                pass

            try:
                # Url
                with ref.inner.fspath.open() as f:
                    nb = nbformat.read(f, as_version=4)
                break
            except:

                pass

            try:
                # File name
                with open(ref) as r:
                    nb = nbformat.read(f, as_version=4)
                break
            except:
                pass

            raise SourceError("Can't handle ref of type {}".format(type(ref)))

        return nb

    def _iter_lines(self, text, row_cb):
        import re

        for line in (text or '').splitlines():
            if re.match(r'^\s*#', line):  # Skip comments
                continue

            # Special handling for ====, which implies a section:
            #   ==== Schema
            # is also
            #   Section: Schema

            if line.startswith('===='):
                line = re.sub(r'^=*','Section:', line.strip())

            row = [e.strip() for e in line.split(':', 1)]

            # Pipe characters separate columns, so they get escaped when they are not seperators
            if len(row) > 1:
                row = [row[0]] + [ e.replace('\|','|') for e in re.split(r'(?<!\\)\|', row[1]) ]

            yield row

            yield from row_cb(row)

    def __iter__(self):

        nb = self._open()

        yield ['Declare', 'metatab-latest']
        yield ['Title', get_cell_source(nb, 'Title').strip('#').strip()]
        yield ['Description', get_cell_source(nb, 'Description') ]

        def row_cb(row):

            if row[0] == 'Section' and row[1] == 'Documentation':
                yield ('Readme', get_cell_source(nb, 'readme'))

        for tag in ['metadata', 'resources', 'schema']:
            yield from self._iter_lines(get_cell_source(nb, tag), row_cb)
